import os
import sys
import pandas

import numpy as np
import matplotlib.pyplot as plt

from lmfit import fit_report
from scipy import stats
from monashspa.common.fitting import linear_fit, get_fit_parameters, make_lmfit_model, model_fit
from monashspa.common.figures import savefig

# Read the data from file
data = None
try:
    fname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'PHS2062_pendulum_data.csv')
    data = pandas.read_csv(fname, usecols=[0,1,2]).to_numpy()
except FileNotFoundError:
    print("ERROR: Failed to read CSV file with data. It should be in the same folder as the Python file.")
if type(data)=="NoneType":
    sys.exit(1)

period = data[:,0]
u_period = data[:,1]
count = data[:,2]


fig, ax = plt.subplots(1)
ax.errorbar(count, period, yerr=u_period, marker="x", linestyle="None", color="black",label="experimental data")
ax.set(xlabel = 'count', ylabel="Period (s)")
fig.suptitle('Pendulum period')
savefig('pendulum_data.png')



# Let us assume that the period is constant

pendulum_model = make_lmfit_model("T", allow_constant_model=True)
pendulum_params = pendulum_model.make_params(T=2.0)
pendulum_fit_results = model_fit(pendulum_model, pendulum_params, count, period, u_y=u_period)

pendulum_fit = pendulum_fit_results.best_fit
u_pendulum_fit = pendulum_fit_results.eval_uncertainty(sigma=1)
pendulum_fit_parameters = get_fit_parameters(pendulum_fit_results)
pvalue = 1.0 - stats.chi2.cdf(pendulum_fit_results.chisqr,
                              pendulum_fit_results.ndata-pendulum_fit_results.nvarys)

print("""
[[Simple pendulum]]
===================
  p-value       = {pvalue:.2E}
""".format(pars=pendulum_fit_parameters, f=pendulum_fit_results, pvalue=pvalue))
print(pendulum_fit_results.fit_report())

pendulum_fit = np.full(len(count), pendulum_fit)
u_pendulum_fit = np.full(len(count), u_pendulum_fit[0])

fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(count, period, yerr=u_period, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(count, pendulum_fit, marker="None", linestyle="-", color="black",label="linear fit")
ax1.fill_between(count, pendulum_fit-u_pendulum_fit,pendulum_fit+u_pendulum_fit,
                 color="lightgrey",label="uncertainty in linear fit")
ax1.legend(bbox_to_anchor=(0.5,1))
ax1.set(ylabel="Period (s)")
pull = (period-pendulum_fit)/u_period
ax2.plot(count, pull, marker="*", linestyle="None", color="black")
xrange = [np.min(count), np.max(count)]
ax2.plot(xrange, [0,0], marker="None", linestyle="-", color="grey")
ax2.plot(xrange, [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot(xrange, [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="Count", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit to a constant period')
savefig('pendulum_constant_fit.png')



# Let us assume a linear change of period

pendulum_model = make_lmfit_model("T*(1.0+a*x)")
pendulum_params = pendulum_model.make_params(T=2.0, a=0)
pendulum_fit_results = model_fit(pendulum_model, pendulum_params, count, period, u_y=u_period)

pendulum_fit = pendulum_fit_results.best_fit
u_pendulum_fit = pendulum_fit_results.eval_uncertainty(sigma=1)
pendulum_fit_parameters = get_fit_parameters(pendulum_fit_results)
pvalue = 1.0 - stats.chi2.cdf(pendulum_fit_results.chisqr,
                              pendulum_fit_results.ndata-pendulum_fit_results.nvarys)

print("""
[[Linear change of period]]
===================
  p-value       = {pvalue:.2E}
""".format(pars=pendulum_fit_parameters, f=pendulum_fit_results, pvalue=pvalue))
print(pendulum_fit_results.fit_report())

fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(count, period, yerr=u_period, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(count, pendulum_fit, marker="None", linestyle="-", color="black",label="linear fit")
ax1.fill_between(count, pendulum_fit-u_pendulum_fit,pendulum_fit+u_pendulum_fit,
                 color="lightgrey",label="uncertainty in linear fit")
ax1.legend(bbox_to_anchor=(0.5,1))
ax1.set(ylabel="Period (s)")
pull = (period-pendulum_fit)/u_period
ax2.plot(count, pull, marker="*", linestyle="None", color="black")
xrange = [np.min(count), np.max(count)]
ax2.plot(xrange, [0,0], marker="None", linestyle="-", color="grey")
ax2.plot(xrange, [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot(xrange, [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="Count", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with a linear term')
savefig('linear_pendulum.png')


# Let us assume an exponential change of period

pendulum_model = make_lmfit_model("T*(1.0+a*exp(-2*k*x))")
pendulum_params = pendulum_model.make_params(T=2.0, k=0)
pendulum_fit_results = model_fit(pendulum_model, pendulum_params, count, period, u_y=u_period)

pendulum_fit = pendulum_fit_results.best_fit
u_pendulum_fit = pendulum_fit_results.eval_uncertainty(sigma=1)
pendulum_fit_parameters = get_fit_parameters(pendulum_fit_results)
pvalue = 1.0 - stats.chi2.cdf(pendulum_fit_results.chisqr,
                              pendulum_fit_results.ndata-pendulum_fit_results.nvarys)

print("""
[[Exponential change of period]]
===================
  p-value       = {pvalue:.2E}
""".format(pars=pendulum_fit_parameters, f=pendulum_fit_results, pvalue=pvalue))
print(pendulum_fit_results.fit_report())

fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(count, period, yerr=u_period, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(count, pendulum_fit, marker="None", linestyle="-", color="black",label="linear fit")
ax1.fill_between(count, pendulum_fit-u_pendulum_fit,pendulum_fit+u_pendulum_fit,
                 color="lightgrey",label="uncertainty in linear fit")
ax1.legend(bbox_to_anchor=(0.5,1))
ax1.set(ylabel="Period (s)")
pull = (period-pendulum_fit)/u_period
ax2.plot(count, pull, marker="*", linestyle="None", color="black")
xrange = [np.min(count), np.max(count)]
ax2.plot(xrange, [0,0], marker="None", linestyle="-", color="grey")
ax2.plot(xrange, [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot(xrange, [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="Count", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with an exponential term')
savefig('exponential_pendulum.png')




plt.show()



# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import pandas

def read_csv(filepath, *args, **kwargs):
    """ A simple wrapper around the pandas csv reader.

    This function wraps the :py:func:`pandas.read_csv` function and returns
    a numpy array. Any optional argument for the pandas function can be used 
    here.
    
    Examples
    ^^^^^^^^
    Standard csv files with a header row can be read using 
    :code:`read_csv(filepath)`.

    If your csv file has no header row, use: 
    :code:`read_csv(filepath, header=None)`.

    If your file is delimited by something other than a comma, you can specify
    the separator with the optional argument :code:`sep`. For example, 
    :code:`sep='|'` if your file is delimited by the pipe symbol.
    
    There are many options that may be useful, such as those that parse dates,
    ignore columns, or convert certain values to 'nan's.
    See the pandas documentation for :py:func:`pandas.read_csv` to see all 
    optional arguments.

    Arguments:
        filepath: The path to the file to read

    Keyword Arguments:
        **kwargs: See the pandas documentation for :py:func:`pandas.read_csv`

    Returns:
        A numpy array. This array will be 1D if the csv file contains only a
        single row or column of data. Otherwise the array will be 2D.
    
    """
    df = pandas.read_csv(filepath, *args, **kwargs)
    arr = df.to_numpy()

    # convert array to 1D if it only has one row or column
    if arr.ndim > 1 and 1 in arr.shape:
        return arr.flatten()
    else:
        return arr


import random

class Particle:
    '''Base class for particles'''

    def __init__(self, type, z, energy, ionise, cutoff):
        self.type = type
        self.z = z
        self.energy = energy
        self.ionise = ionise
        self.cutoff = cutoff

    def move(self, step):
        self.z += step

    def interact(self):
        '''This should implement the model for interaction.
        The base class particle doesn't interact at all'''
        return [particle]

    def __str__(self):
        return f'{self.type:10} z:{self.z:.3f} E:{self.energy:.3f}'


class Electron(Particle):

    def __init__(self, z, energy):
        super(Electron, self).__init__('elec', z, energy, True, 0.01)

    def interact(self):
        '''An electron radiates a photon. Make the energy split evenly.'''
        particles = []
        if self.energy > self.cutoff:
            split = random.random()
            particles = [Electron(self.z, split*self.energy), Photon(self.z, (1.0-split)*self.energy)]
        return particles


class Photon(Particle):

    def __init__(self, z, energy):
        super(Photon, self).__init__('phot', z, energy, False, 0.01)

    def interact(self):
        '''A photon splits into an electron and a positron. Make the energy split evenly.'''
        particles = []
        if self.energy > self.cutoff:
            split = random.random()
            particles = [Electron(self.z, split*self.energy), Electron(self.z, (1.0-split)*self.energy)]
        return particles


class Muon(Particle):

    def __init__(self, z, energy):
        super(Muon, self).__init__('muon', z, energy, True, 0.01)



import random
import os
import csv
import math

import numpy as np
import matplotlib.pyplot as plt

from scipy.integrate import solve_ivp
from monashspa.common.figures import savefig


# Gas data created according to van der Waals law

a_vdW = 20*1.370 # 20 * Nitrogen
b_vdW = 50*0.0387 # 50 * Nitrogen

n_gas = 0.040 # moles
R_gas = 0.08314 # L bar K^-1 mol^-1

volume = np.arange(start=0.3, stop=2.0, step=0.05) # L
volume = volume + np.random.normal(0.0, 0.04, len(volume))
pressure = 1.013 # bar

temp_vdW = ( pressure*volume + a_vdW*n_gas*n_gas*(1-n_gas*b_vdW) - pressure*n_gas*b_vdW ) / (n_gas*R_gas)
temp_ideal = ( pressure*volume ) / (n_gas*R_gas)
volume_data = volume + np.random.normal(0.0, 0.01, len(volume))
u_volume_data = np.full((len(temp_vdW)), 0.01)

fname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'PHS2062_gas_data.csv')
with open(fname, 'w', newline='') as csvfile:
    writer = csv.writer(csvfile, delimiter=',', quotechar='\'', quoting=csv.QUOTE_MINIMAL)
    writer.writerow( ['volume (L)', 'Volume uncertainty (L)', 'Temperature (K)'] )
    for v, u_v, t in zip(volume_data, u_volume_data, temp_vdW):
        writer.writerow(["{0:>7.3f}   ".format(v), "{0:>14.3f}        ".format(u_v), "{0:>8.1f}".format(t)])



volume = np.arange(start=1.0, stop=1.25, step=0.03) # L
volume = volume + np.random.normal(0.0, 0.015, len(volume))
pressure = 1.013 # bar

temp_vdW = ( pressure*volume + a_vdW*n_gas*n_gas*(1-n_gas*b_vdW) - pressure*n_gas*b_vdW ) / (n_gas*R_gas)
temp_ideal = ( pressure*volume ) / (n_gas*R_gas)
volume_data = volume + np.random.normal(0.0, 0.03, len(volume))
u_volume_data = np.full((len(temp_vdW)), 0.01)

fname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'PHS2062_gas_short_data.csv')
with open(fname, 'w', newline='') as csvfile:
    writer = csv.writer(csvfile, delimiter=',', quotechar='\'', quoting=csv.QUOTE_MINIMAL)
    writer.writerow( ['volume (L)', 'Volume uncertainty (L)', 'Temperature (K)'] )
    for v, u_v, t in zip(volume_data, u_volume_data, temp_vdW):
        writer.writerow(["{0:>7.3f}   ".format(v), "{0:>14.3f}        ".format(u_v), "{0:>8.1f}".format(t)])


volume = np.arange(start=1.0, stop=1.25, step=0.02) # L
volume = volume + np.random.normal(0.0, 0.015, len(volume))
pressure = 1.013 # bar

temp_vdW = ( pressure*volume + a_vdW*n_gas*n_gas*(1-n_gas*b_vdW) - pressure*n_gas*b_vdW ) / (n_gas*R_gas)
temp_ideal = ( pressure*volume ) / (n_gas*R_gas)
volume_data = volume + np.random.normal(0.0, 0.03, len(volume))
u_volume_data = np.full((len(temp_vdW)), 0.03)

fname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'PHS2062_gas_short_2_data.csv')
with open(fname, 'w', newline='') as csvfile:
    writer = csv.writer(csvfile, delimiter=',', quotechar='\'', quoting=csv.QUOTE_MINIMAL)
    writer.writerow( ['volume (L)', 'Volume uncertainty (L)', 'Temperature (K)'] )
    for v, u_v, t in zip(volume_data, u_volume_data, temp_vdW):
        writer.writerow(["{0:>7.3f}   ".format(v), "{0:>14.3f}        ".format(u_v), "{0:>8.1f}".format(t)])


        


# Measured period of large angle pendulum with friction

def pendulum(t, y):
    return [ y[1] , -9.82 /50. * math.sin(y[0]) - 0.025 * y[1] ]

def zerocrossing(t, y):
    return y[0]
zerocrossing.direction=1.0

sol = solve_ivp(pendulum, [0, 300], [math.pi/2., 0.0], events=zerocrossing, max_step=0.1)

# Assume that acuracy of timings are better when pendulum is moving fast
n_crossings = len(sol.t_events[0])
timeuncertainty = np.linspace(0.05, 0.3, n_crossings)
crossingtimes = sol.t_events[0] + np.random.normal(0.0, timeuncertainty)
periods = crossingtimes[1:-1] - crossingtimes[0:-2]
u_periods = math.sqrt(2)*timeuncertainty[:-2]
count = np.linspace(0,n_crossings-2,n_crossings-1)

fname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'PHS2062_pendulum_data.csv')
with open(fname, 'w', newline='') as csvfile:
    writer = csv.writer(csvfile, delimiter=',', quotechar='\'', quoting=csv.QUOTE_MINIMAL)
    writer.writerow( ['Period (s) ', 'Period uncertainty ', '  count'] )
    for p, u_p, c in zip(periods, u_periods, count):
        writer.writerow(["{0:>8.3f}   ".format(p), "{0:>11.3f}        ".format(u_p), "{0:>5.1f}".format(c)])

plt.figure(1)
plt.plot(sol.t, sol.y[0], marker=None, linestyle="-", color="red",label="amplitude")
plt.xlabel("time (s)")
plt.ylabel("angle (rad)")
plt.title('pendulum')
savefig("pendulum_angle.png")
plt.show()



# Copyright 2019 School of Physics & Astronomy, Monash University),
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import pandas

def read_picoscope_csv(filepath):
    """Reads the csv file saved from the PicoScope software
    
    .. note:: The picoscope software will save overrange values in the csv file
              as either "Infinity" or the unicode symbol for infinity (which 
              will show up in some software as several random characters). This
              Python function converts those instances to :code:`np.nan`.

    .. note:: This function attempts to return values in S.I. units without a 
              prefix. For example, if the csv file contains time in 
              milliseconds, this function will attempt to detect that and return 
              the data in units of seconds. If the function fails to make this 
              conversion, a warning message will be printed.

    Arguments:
        filepath: A string containing the path to the csv file

    Returns:
        A tuple :code:`(time, CHA, CHB)` where each element is a 1D numpy 
        array containing an array of time 
        points (x-axis of the oscilloscope), and the y-values for channels A
        and B respectively. The units of these should be seconds, Volts and 
        Volts respectively unless otherwise stated via a warning message 
        printed to your terminal.
    """

    # read the data (skip the second line, and only use first 3 columns)
    df = pandas.read_csv(filepath, skiprows=[1,2], usecols=[0,1,2], na_values=['Infinity', '-Infinity', '\u221e', '-\u221e'])

    time = df[df.columns[0]].values
    CHA = df[df.columns[1]].values
    CHB = df[df.columns[2]].values

    # now read the units
    df = pandas.read_csv(filepath, nrows=2, usecols=[0,1,2])
    
    time_unit = df[df.columns[0]].values[0]
    CHA_unit = df[df.columns[1]].values[0]
    CHB_unit = df[df.columns[2]].values[0]

    if time_unit == '(ms)':
        time = time * 1e-3
    elif time_unit == '(s)':
        pass
    elif time_unit == '(us)':
        time = time * 1e-6
    else:
        print('Warning: Unknown unit for time. Do not assume time is in seconds. It is in {}'.format(time_unit))

    if CHA_unit == '(mV)':
        CHA = CHA * 1e-3
    elif CHA_unit == '(V)':
        pass
    elif CHA_unit == ('uV'):
        CHA = CHA * 1e-6
    else:
        print('Warning: Unknown unit for CHA. Do not assume CHA is in Volts. It is in {}'.format(CHA_unit))

    if CHB_unit == '(mV)':
        CHB = CHB * 1e-3
    elif CHB_unit == '(V)':
        pass
    elif CHB_unit == ('uV'):
        CHB = CHB * 1e-6
    else:
        print('Warning: Unknown unit for CHB. Do not assume CHB is in Volts. It is in {}'.format(CHB_unit))

    return time, CHA, CHB


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .testing_helpers import compare_dictionary

def test_basic_linear_fit():
    """Basic test of linear fit without uncertainties"""

    ### Get results ###
    from monashspa.common.fitting import linear_fit, get_fit_parameters

    data = [
        [1,	2.97958155926148, 0.435952175570949],
        [2,	4.84845692653319, 0.467234424802873],
        [3,	6.44561871392239, 0.740203944022575],
        [4,	8.72544337668039, 0.509084442407005],
        [5,	10.7326357174404, 0.557439432218125],
        [6,	12.3202409005262, 0.0236735417569119],
    ]
    data = np.array(data)

    fit_result = linear_fit(data[:,0], data[:,1])
    results = get_fit_parameters(fit_result)

    ### Expected results ###
    expected_results = {
        'slope': 1.903875935,
        'u_slope': 0.045590672,
        'intercept': 1.011763758,
        'u_intercept': 0.177550159,
    }
    precision = 1e-3

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success


def test_linear_fit_with_uncertainties():
    """Basic test of linear fit with uncertainties"""

    ### Get results ###
    from monashspa.common.fitting import linear_fit, get_fit_parameters

    data = [
        [1,	2.97958155926148, 0.435952175570949],
        [2,	4.84845692653319, 0.467234424802873],
        [3,	6.44561871392239, 0.740203944022575],
        [4,	8.72544337668039, 0.509084442407005],
        [5,	10.7326357174404, 0.557439432218125],
        [6,	12.3202409005262, 0.0236735417569119],
    ]
    data = np.array(data)

    fit_result = linear_fit(data[:,0], data[:,1], u_y=data[:,2])
    results = get_fit_parameters(fit_result)

    ### Expected results (from WFIT) ###
    expected_results = {
        'slope': 1.866046353,
        'u_slope': 0.064841588,
        'intercept': 1.124423955,
        'u_intercept': 0.387570521,
    }
    precision = 1e-3

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def test_failed_fit_1():
    """A test of the exception that should be raised if you don't use the independent variable"""
    ### Get results ###
    import monashspa.PHS2061 as spa
    from monashspa.common.fitting import MonashSPAFittingException

    # load the data
    data = spa.fitting_tutorial.data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model1 = spa.make_lmfit_model("A_0*exp(-l*x)")
    success = False
    try:
        model2 = spa.make_lmfit_model("A_0")
    except MonashSPAFittingException as e:
        if str(e).startswith('You have not used the independent variable'):
            success=True

    return success

def test_bypass_failed_fit_1():
    """A test of suppressing the exception that should be raised if you don't use the independent variable"""
    ### Get results ###
    import monashspa.PHS2061 as spa

    # load the data
    data = spa.fitting_tutorial.data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model1 = spa.make_lmfit_model("exp(-l*x)")
    model2 = spa.make_lmfit_model("A_0", allow_constant_model=True)
    model = model1*model2
    params = model.make_params(A_0=30, l=.005)
    params.add('halflife', expr="log(2)/l")
    fit_results = spa.model_fit(model, params, x=t, y=A, u_y=u_A)
    results = spa.get_fit_parameters(fit_results)
    
    ### Expected results ###
    expected_results = {
        'A_0': 20.573035990441486,
        'u_A_0': 0.6181473325960677,
        'l': 0.005004779941925933,
        'u_l': 0.00015840653659960648,
        'halflife': 138.49703455557113,
        'u_halflife': 4.38357646646529,
    }
    precision = 1e-3
    
    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def test_failed_fit_2():
    """A test of not specifying a good initial guess for a parameter"""
    ### Get results ###
    import monashspa.PHS2061 as spa
    from monashspa.common.fitting import MonashSPAFittingException

    # load the data
    data = spa.fitting_tutorial.data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model = spa.make_lmfit_model("A_0*exp(-l*x)")
    params = model.make_params(A_0=30) # l will be "1" by default which is a bad guess
    params.add('halflife', expr="log(2)/l")
    success = False
    try:
        fit_results = spa.model_fit(model, params, x=t, y=A, u_y=u_A)
    except MonashSPAFittingException as e:
        if str(e).endswith('The fit failed. This is usually either because (a) you did not provide sufficient guesses for the model parameters, (b) you did not correctly specify the independent variable in your model (by default it must be "x"), or (c) the data you are fitting to contains NaN values.'):
            success=True
    
    return success

def do_tests():
    tests = [test_basic_linear_fit, test_linear_fit_with_uncertainties, test_failed_fit_1, test_bypass_failed_fit_1, test_failed_fit_2]
    failed_tests = []
    print('Running common fitting tests...')

    for testfn in tests:
        print('    Running test "{test_name}":'.format(test_name=testfn.__doc__))
        result = testfn()
        print('        Result: {result}'.format(result='success' if result else 'failure'))

        if not result:
            failed_tests.append(testfn)

    if failed_tests:
        print('')
        print('    There were {num_failures:d} failed common fitting tests'.format(num_failures=len(failed_tests)))
        
    print('')

    return failed_tests

if __name__ == "__main__":
    do_tests()



import random

class Layer:
    '''Defines an individual layer of a calorimeter. The properties of the layer are
    name, its material given as X0 per cm, the thickness, the response measuring the
    level of ionisation (in arbitrary units, zero for passive layer). The layer can
    keep track of the ionisation in it.'''

    def __init__(self, name, material, thickness, response=1.0):
        self._name = name
        self._material = material
        self._thickness = thickness
        self._yield = response
        self._ionisation = 0

    def ionise(self, particle, step):
        '''Records the ionisation in each layer from a particle going a certain length.'''
        if particle.ionise:
            self._ionisation += self._yield*step

    def interact(self, particle, step):
        '''Let a particle interact (bremsstrahlung or pair production). The interaction
        length is assumed to be the same for electrons and photons.'''
        material = self._material*step
        particles = [particle]
        if random.random() < material:
            particles = particle.interact()

        return particles

    def __str__(self):
        return f'{self._name:10} {self._material:.3f} {self._thickness:.2f} cm {self._ionisation:.3f}'



#!/usr/bin/env python
# coding: utf-8

# # $D^0$ lifetime task

# ## Introduction

# In this assignment we will try to measure the lifetime of the $D^0$ meson.
# This meson contains a charm quark and an anti up quark $(c\bar{u})$. When it decays, the most likely process is the charm quark decaying into a strange quark plus a up and anti-down $(c\rightarrow su\bar{d})$. This is mediated by the weak force, and means the meson survives longer than particles that can decay via the electromagnetic or strong forces. 
# 
# 
# The data sample we will use for this assignment was taken by the LHCb experiment using collisions between two beams of protons. We will only look at a very small fraction of the total sample. When the $D^0$ mesons are produced at the LHC they will be traveling relativistically. As a result of time-dilation the distance they typically travel before decaying can be as much as a few millimeters even though their life time is on the order of picoseconds. Most particles produced in proton-proton collisions are emitted directly from the point of the collision (primary vertex), therefore we can use the position of the $D^{0}$ decay (secondary vertex) to help disentagle them from the many other particles produced in the collision.
# 
# Notes:
# - All momentums are given in units of ${\rm MeV}$ and distances in $\rm mm$.
# 

# ## Part 1: Reading in the data

# To begin with we will download the data set and access the information we will need.
# This file contains around 100,000 $D^0\rightarrow K^{-} \pi^{+}$ decays.
# 
# The file contains:
# - the 3 components of momentum for the kaon and pion,
# - the position of the proton-proton collision (primary vertex),
# - the position of the $D^0$ decay (secondary vertex).


import monashspa
import uproot4 as uproot
import matplotlib.pyplot as plt
import numpy as np
import math
from monashspa.common.download_gdrive import download_file_from_google_drive

## define constants
const_c  = 2.99792e8 * 1e3 * 1e-12 ## in units mm/ps



## Download the data
download_file_from_google_drive('1sJI2nsJGmCu52xvpqCaijYQlDyqGA2b7','Data_small.root')

## Access the information we need
file = uproot.open("Data_small.root")
tree = file["RS_LTUNB"]

## momentum
K_PX  = tree["P1_PX"].array(library="np")
Pi_PX = tree["P2_PX"].array(library="np")

K_PY  = tree["P1_PY"].array(library="np")
Pi_PY = tree["P2_PY"].array(library="np")

K_PZ  = tree["P1_PZ"].array(library="np")
Pi_PZ = tree["P2_PZ"].array(library="np")

## secondary vertex position
D0_X = tree["D0_ENDVERTEX_X"].array(library="np")
D0_Y = tree["D0_ENDVERTEX_Y"].array(library="np")
D0_Z = tree["D0_ENDVERTEX_Z"].array(library="np")

## primary vertex position
PV_X = tree["D0_OWNPV_X"].array(library="np")
PV_Y = tree["D0_OWNPV_Y"].array(library="np")
PV_Z = tree["D0_OWNPV_Z"].array(library="np")

## Calculate transverse momentum
K_PT  = np.sqrt(K_PX**2 +K_PY**2)
Pi_PT = np.sqrt(Pi_PX**2+Pi_PY**2)



## Lets try plotting the momentum of the kaon and pion
fig, ax = plt.subplots(2, 2,figsize=(8,5))
ax[0,0].hist(K_PT,bins=100,label="K_PT")
ax[0,0].set_xlabel('$p_{T}(K^-)$ [MeV]')
ax[0,0].set_ylabel('Candidates')

ax[0,1].hist(K_PZ,bins=100)
ax[0,1].set_xlabel('$p_{z}(K^-)$ [MeV]')
ax[0,1].set_ylabel('Candidates')

ax[1,0].hist(Pi_PT,bins=100)
ax[1,0].set_xlabel('$p_{T}(\pi^+)$ [MeV]')
ax[1,0].set_ylabel('Candidates')

ax[1,1].hist(Pi_PZ,bins=100)
ax[1,1].set_xlabel('$p_{z}(\pi^+)$ [MeV]')
ax[1,1].set_ylabel('Candidates')
plt.tight_layout()
plt.show()



## plot the position of the D0 production and decay verticies
fig, ax = plt.subplots(2, 3, figsize=(10,5))
ax[0,0].hist(D0_X,bins=100)
ax[0,0].set_xlabel('$x_{decay}$ [mm]')
ax[0,0].set_ylabel('Candidates')

ax[0,1].hist(D0_Y,bins=100)
ax[0,1].set_xlabel('$y_{decay}$ [mm]')
ax[0,1].set_ylabel('Candidates')

ax[0,2].hist(D0_Z,bins=100)
ax[0,2].set_xlabel('$z_{decay}$ [mm]')
ax[0,2].set_ylabel('Candidates')

ax[1,0].hist(PV_X,bins=100)
ax[1,0].set_xlabel('$x_{production}$ [mm]')
ax[1,0].set_ylabel('Candidates')

ax[1,1].hist(PV_Y,bins=100)
ax[1,1].set_xlabel('$y_{production}$ [mm]')
ax[1,1].set_ylabel('Candidates')

ax[1,2].hist(PV_Z,bins=100)
ax[1,2].set_xlabel('$z_{production}$ [mm]')
ax[1,2].set_ylabel('Candidates')

plt.tight_layout()
plt.show()


# ## Part 2: Creating the information we'll need
# 
# The data set contains the three components of momentum  of the kaon and pion, and the positions of their vertex and the proton-proton interaction vertex. From this we will need to create the information we need: the mass and decay time of the $D^0$ candidate.  
# 
# ### Part 2a: $D^0$ mass
# The data contains the 3-momenta of the kaon and pion:
# $$
# \vec{\mathbf{p}} = (p_{x},p_{y},p_{z})
# $$
# We need the 4-momenta, which means adding the energy of the particles as well. At the LHCb experiment we don't measure the particle energy directly. Instead we use particle identification detectors to determine what type of particle the various the long-lived particles are (e.g. kaons or pions) and use the externally measured values of their masses and our measured momentums to compute the energy.
# Values of the pion and kaon masses have been measured elsewhere:
# - $m(\pi) = 139.6\,{\rm MeV}/c^{2}$
# - $m(K) = 493.7\,{\rm MeV}/c^{2}$ 
# 
# **Task 1:** *[2 marks]* Create new variables called `K_E` and `Pi_E` in which you calculate the pion and kaon energies from the 3-momenta and their known masses.


# Pi_E = 
# K_E = 


# Now we have the pion and kaon energies we have the complete 4-momentum for the kaon and pion
# 
# $$ 
# \mathbf{p}_{\pi} = \left[E_\pi,p_{x}(\pi),p_{y}(\pi),p_{z}(\pi)\right]
# $$
# $$
# \mathbf{p}_{K} = \left[E(K),p_{x}(K),p_{y}(K),p_{z}(K)\right]
# $$
# we can use these to calculate the mass of the $D^0$ meson candidate. 
# 
# **Task 2:** *[2 marks]* Calculate the $D^0$ mass using the all of the components of the kaon and pion 4-momentum. 


## D0_M = 


# Our sample contains both real $D^0\to K \pi$ decays as well as backgrounds from combinations of a kaon and pion that didn't originate from a $D^0$ meson. A loose set of requirements have already been applied to the sample we are analysing to reduce the number of background candidates, whilst keeping most of the real $D^0\to K \pi$ decays.
# These two different contributions will be easy to see if you've calculated the $D^0$ mass correctly. Hint: the mass of the $D^0$ meson is $1864.84\pm0.05\,{\rm MeV}/c^2$. 
# 
# **Task 3:** *[2 marks]* Create a histogram of the $D^0$ mass.



## Plot D0_M histogram


# ### Part 2b: $D^0$ lifetime
# To calculate the lifetime we first need to determine the distance between the proton-proton interaction vertex ($D^0$ production vertex, or primary vertex) and the $D^0$ decay vertex (or secondary vertex). The 3-vector between these points is given by:
# $$
# \Delta\vec{x} = \vec{x}_{\rm decay} - \vec{x}_{\rm production}.
# $$
# Note, both the production and decay positions are given in units of $\rm mm$. 
# 
# **Task 4:** *[4 marks]* Calculate and plot a histgram of the absolute distance $|\Delta\vec{x}|$ between the primary vertex (proton-proton collision) and secondary vertex ($D^0$ decay).


# D0_L = 



## Plot D0_L histogram


# In the lab frame, a particle travelling at velocity $\vec{\mathbf{v}}$ for a time $t$ will travel a distance $\Delta \vec{\mathbf{x}}$:
# $$
# \Delta\vec{\mathbf{x}} = \vec{\mathbf{v}}t
# $$
# 
# We must account for special relativity when trying to determine the lifetime of the $D^0$ meson. We can rewrite this equation using relations between the momentum and velocity, and between lab frame time and proper time:
# 
# $$
# \vec{\mathbf{p}} = \gamma m \vec{\mathbf{v}}
# $$
# 
# $$
# t = \gamma \tau.
# $$
# 
# **Task 5** *[2 marks]* Derive an experession for the decay time $\tau$ in terms of $\Delta\vec{\mathbf{x}}$, $\vec{\mathbf{p}}$, and $m$.
# 
# If we use $\vec{\mathbf{p}}$ in units of $[{\rm MeV}]$, $m$ in units of $[{\rm MeV}]$ and $\Delta\vec{\mathbf{x}}$ in units of $[\rm mm]$, the proper time $\tau$ will be calculated in units of $[\rm mm]$ (i.e. it's actually a length. This is because we are using natural units for the momentum and mass but SI units for the positions). 
# 
# We can convert $[\rm mm]$ into $[\rm ps]$ by dividing  by the speed of light in units of $[\rm mm \,ps^{-1}]$ (given at the top of the script).

# **Task 6** *[5 marks]* Calculate and plot a histgram of the proper time $\tau$ in units of picoseconds. *Hint: you'll need to create arrays containing the $D^0$ total momentum and $p_{x}$, $p_{y}$ and $p_{z}$ components.*


# Calculate tau [ps]
# Plot histogram of tau


# ### Part 3: Background subtraction

# You should be able to see from your histogram of the $D^0$ mass that there is a component that is flat as a function of $m_{D^0}$ and a contribution that looks like a Gaussian. The Gaussian is our signal and the flat contribution is from background. This can be random combinations of kaons and pions that happen to have a $D^0$ mass within the mass window we've selected. These background decays may not have the same decay time distribution as our signal, so we must subtract the distribution of the background from our signal when plotting histograms of the decay time. 
# 
# There are a number of ways to perform background subtraction, we will use a simple method, as in our case the background distribution is very flat. 
# - First, we will define a signal region. This could be $\pm 20\,{\rm MeV}$ around the known $D^0$ mass $1844.84<m_{D^0}<1884.84\,{\rm MeV}$.
# 
# - Second, we will define *two* background-only regions, each $ 20\,{\rm MeV}$ wide.
#     - e.g. $1820<m_{D^0}<1840\,{\rm MeV}$ and  $1890<m_{D^0}<1910\,{\rm MeV}$
# 
# - Then if we plot a histogram of data in the signal region, and subtract a histogram of data in the two background-only regions we should end up with just the signal contribution. (This works because the width of the two background regions sums up to the same as the signal region and should cancel out the corresponding background decays) 
# 
# To plot a histogram of only a subset of the data we can use a mask. 
# An example is shown below for momentum.
# 


## Example of using masks to only plot a subset of data

if False:
    # Define some signal and background regions
    # Note- these are just for demonstration purposes
    signal_mask     = (D0_PX>-10000)&(D0_PX<10000)
    background_mask = (D0_PX<-10000)|(D0_PX>10000)

    fig, ax = plt.subplots(1,2,figsize=(10,5))

    ## We must use the same range for both so the correct values are subtracted
    hist_sig = ax[0].hist(D0_PY[signal_mask] ,    bins=50, range = (-20000,20000)) 
    hist_bkg = ax[0].hist(D0_PY[background_mask] ,bins=50, range = (-20000,20000))
    ax[0].set_yscale('log')


    ## Define a function to create the difference of two histograms
    def subtractHitograms(hist_sig,hist_bkg):
        val_of_bins_sig, edges_of_bins_sig, _ = hist_sig
        val_of_bins_bkg, edges_of_bins_bkg, _ = hist_bkg

        difference = val_of_bins_sig - val_of_bins_bkg
        ## Using possion statistics the error on each bin is sqrt(N)
        ## So we add sqrt(N_sig) and sqrt(N_bkg) in quaderature 
        error      = np.sqrt(val_of_bins_sig + val_of_bins_bkg,where=(val_of_bins_sig+val_of_bins_bkg >= 0))
        bincenter  = 0.5 * (edges_of_bins_sig[1:] + edges_of_bins_sig[:-1])
        return difference,error,bincenter

    ## Plot the difference between two histograms
    difference,error,bincenter = subtractHitograms(hist_sig,hist_bkg)
    ax[1].errorbar(bincenter, difference, yerr=error, color='r')
    ax[1].set_yscale('log')
    plt.show()


# 
# **Task 7** *[5 marks]* Plot histograms of the $D^0$ decay time in the signal and background-only regions. Take the difference between the histograms to create the background-subtracted decaytime distribution. 


## Define signal and background masks

## Plot total, background and background-subtracted tau distributions  


# ### Part 4: Fit the decay time distribution to determine the $D^0$ lifetime 
# 
# The $D^0$ lifetime is a similar concept to the half life of a radioactive isotope. After one lifetime, on average, a certain fraction of the original sample will have decayed. Like radioactive decay, this means the decay time distribution ($\tau$) will follow an falling exponential, with the exponential slope given by the lifetime $\tau_{D^0}$
# $$
# \mathcal{P}(\tau) = N \exp{\left(-\frac{\tau}{\tau_{D^0}}\right)},
# $$
# where $N$ is a normalisation factor.
# We want to fit our background-subtracted distribution to determine the lifetime.
# 
# An example fit is given below. 
#  
# 
# Think about:
# - What range of decay times should you include in you histogram?
# - Why might the distribution not look purely exponential at small decay times? 
# 
# **Task 8** *[5 marks]* Fit your decay time distribution to determine an estimate for the $D^0$ meson lifetime and an error. How does this compare to the known values? Use the [Particle Data Group](https://pdglive.lbl.gov/) value as a reference. What systematic sources of uncertainty might contribute in addition to the statistical uncertainty on your measurement?
# 


from lmfit import fit_report
from scipy import stats
from monashspa.common.fitting import linear_fit, get_fit_parameters, make_lmfit_model, model_fit
from monashspa.common.figures import savefig

if False:
    ## Fill these with the contents of your background-subtracted distrbution
    x_values = []
    y_values = []
    y_errors = []

    ## Define a model to fit and set the parameters
    name        ='Decay time'
    model       = make_lmfit_model("a*sin(x/b)")
    params      = model.make_params(a=1.0, b=1.0)
    fit_results = model_fit(model, params, x_values, y_values, u_y=y_errors)

    # Extract result and print nicely
    fit            = fit_results.best_fit
    u_fit          = fit_results.eval_uncertainty(sigma=1)
    fit_parameters = get_fit_parameters(fit_results)
    pvalue         = 1.0 - stats.chi2.cdf(fit_results.chisqr,
                                      fit_results.ndata-fit_results.nvarys)

    ## Plot the result

    print("""
    [[{name}]]
    =================
      p-value       = {pvalue:.2E}
    """.format(name=name, pvalue=pvalue))
    print(fit_results.fit_report())

    # Create some plots
    fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})

    ax1.errorbar(    x_values, y_values, 
                     yerr=y_errors, 
                     marker="x", 
                     linestyle="None", 
                     color="black",
                     label=name)

    ax1.plot(        x_values, fit, 
                     marker="None", 
                     linestyle="-", 
                     color="black",
                     label="fit to {name}".format(name=name))

    ax1.fill_between(x_values, fit-u_fit, fit+u_fit,
                     color="lightgrey",
                     label="uncertainty in {name}".format(name=name))
    ax1.legend()
    ax1.set(ylabel="Background-subtracted candidates")
    ax1.set_yscale('log')

    ## Plot the pulls
    pull = (y_values-fit)/y_errors
    ax2.plot(x_values, pull, marker="*", linestyle="None", color="black")
    emin = np.min(x_values)
    emax = np.max(x_values)
    ax2.plot([emin, emax], [ 0, 0], marker="None", linestyle="-", color="grey")
    ax2.plot([emin, emax], [ 1, 1], marker="None", linestyle="dashed", color="grey")
    ax2.plot([emin, emax], [-1,-1], marker="None", linestyle="dashed", color="grey")
    ax2.set(xlabel="x", ylabel="Pull")
    scale = 1.1*np.max(np.abs(pull))
    ax2.set_ylim(-scale, scale)
    fig.suptitle('Fit with {name}'.format(name=name))

    plt.show()








import copy
import numpy as np

class Simulation:
    '''A simulation is defined by a calorimeter. Then individual simulation runs can be created by
    running the same particle through the calorimter multiple times.'''
    def __init__(self, calorimeter):
        self._calorimeter = calorimeter

    def simulate(self, particle, number):
        '''Run a individual simulation. The ingoing particle is simulated going
        through the calorimeter "number" times. A 2D array is returned with the
        first axis the ionisation in the individual layers and the second corresponding to each
        new particle.'''
        ionisations = []

        for i in range(number):

            self._calorimeter.reset()
            particles = [copy.copy(particle)]
            step = 0
            while step < 10000:
                next = []
                for p in particles:
                    newparticles = self._calorimeter.step(p, 0.1)
                    next.extend(newparticles)
                particles = next
                step += 1

            ionisations.append(self._calorimeter.ionisations())

        allionisations = np.stack(ionisations, axis=0)
        return allionisations



import requests

def download_file_from_google_drive(id, destination):
    """
    Download a files from Google drive. The file has to shared and made viewable by everybody. The id of the
    file is the long list of random characters in the public link.

    Example: download_file_from_google_drive('1c8FPxYIdRNY8WmB8G4k7p_DSNrWwwfwfD','myfile.txt'
    
    """
    def get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value

        return None

    def save_response_content(response, destination):
        CHUNK_SIZE = 32768

        with open(destination, "wb") as f:
            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)

    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination)    



#!/usr/bin/env python
# coding: utf-8

# ------------------------------------------------------------------------------
# Q3: Fourier analysis
# Part 2: Synthesising an instrument
# ------------------------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt 
import scipy.io.wavfile
# from IPython.display import Audio
from matplotlib.widgets import Cursor

# ------------------------------------------------------------------------------
# We're going to create a set of sin functions. 
# You can specify the amplitudes, frequencies and phases shifts for each of the waves. 
# Try adding more of the harmonics you found when analysing your recordings.
# ------------------------------------------------------------------------------


# Samples per second
sample_rate = 44100

# Duration in seconds
duration_s = 5

## Create a list of sine waves to add together 
## Each one has an amplitude, frequency (Hz) and phase (radians)
wave_properties = [
    (0.4, 100.0, 0.0),
    (0.1, 200.0, np.pi/2.0),
    (0.005, 400.0, 0.0),
]

## Create the waveforms for each sine wave
waveforms = []
sample_number = np.arange(duration_s * sample_rate)
for amp,freq,phase in wave_properties:
    waveforms += [amp*np.sin(2 * np.pi * sample_number * freq / sample_rate + phase)]

# ------------------------------------------------------------------------------
# Now we can plot each of the waves we've created, along with the total sum.
# ------------------------------------------------------------------------------

## Plot a restricted range of the samples
first_sample = 0
last_sample  = 1000

total = np.zeros(duration_s * sample_rate)

fig, (ax0, ax1) = plt.subplots(1, 2)
for wave in waveforms:
    total += wave
    x = sample_number[first_sample:last_sample]
    y = wave[first_sample:last_sample]
    ax0.plot(x,y)
ax0.set_xlabel('Sample number')
ax0.set_ylabel('Amplitude')
ax0.set_title('Individual waves')


x = sample_number[first_sample:last_sample]
y = total[first_sample:last_sample]
ax1.plot(x,y)
ax1.set_xlabel('Sample number')
ax1.set_title('Total wave')

fig.suptitle('Synthesised waveforms')

# save the plot
plt.savefig('part2_sine_waves.png')

## Save the sample and see what it sounds like!
max_amplutide = np.iinfo(np.int32).max
scipy.io.wavfile.write("waveform_unwindowed.wav", sample_rate, (max_amplutide*total).astype(np.int32))
# ------------------------------------------------------------------------------
# To make the wave sound more like the guitar string, we can multiple it be an
# envelope function to mimic how it moves when we pluck the string, and as the 
# energy is disipated over time. 
# ------------------------------------------------------------------------------

## Lets use an exponential function to model the decay
## and 1-exp to model the pluck

decay_time = 1.0 ## seconds 
decay_shape = np.exp(-sample_number/(decay_time*sample_rate) )

attack_time = 0.01 ## seconds 
attack_shape = 1-np.exp(-sample_number/(attack_time*sample_rate) )

windowed = total*decay_shape*attack_shape

## We can plot a reduced range if we want to zoom in 
first_sample = 0
last_sample  = len(sample_number)
# first_sample = 0
# last_sample  = 25000

fig, (ax0,ax1,ax2) = plt.subplots(1, 3)
ax0.plot(sample_number[first_sample:last_sample],decay_shape[first_sample:last_sample])
ax0.set_xlabel('Sample number')
ax0.set_ylabel('Amplitude')
ax0.set_title('Decay function')

ax1.plot(sample_number[first_sample:last_sample],attack_shape[first_sample:last_sample])
ax1.set_xlabel('Sample number')
ax1.set_title('Switch-on function')

ax2.plot(sample_number[first_sample:last_sample],windowed[first_sample:last_sample])
ax2.set_xlabel('Sample number')
ax2.set_title('Total waveform')

fig.suptitle('Windowing waveforms')

plt.savefig('part2_total_waveform.png')

# ------------------------------------------------------------------------------
# Now we can save our function as an audio file and listen to it.
# ------------------------------------------------------------------------------

## Save the sample and see what it sounds like!
scipy.io.wavfile.write("waveform_windowed.wav", sample_rate, (max_amplutide*windowed).astype(np.int32) )

## Either open the file above in an audio player or play directly in the notebook
# Audio(windowed, rate=sample_rate) 

# ------------------------------------------------------------------------------
# We can compare the fourier transform of our newly created wave to the 
# transform of our recording. 
# How similar is it?
# ------------------------------------------------------------------------------


## What does the fourier transform of the waveform look like now?
## Perform a fast fourier transform of the waveform
data_to_transform = windowed
fourier_transform = np.fft.rfft(data_to_transform)

## Take the absolute value
abs_fourier_transform =  np.abs(fourier_transform)

## Create an array with the frequencies for each point in the transform 
freq = np.fft.rfftfreq(data_to_transform.size, d=1./sample_rate)


fig, (ax0) = plt.subplots(1, 1)
ax0.plot(freq,abs_fourier_transform)
ax0.set_xlabel('Frequency (Hz)')
ax0.set_ylabel('Absolute value of FFT')
ax0.set_yscale('log')
ax0.set_title('Fourier transfrom of synthesised waveform')

## Pick which range of frequencies to plot
ax0.set_xlim([0, 1000])
cursor = Cursor(ax0, color='green', linewidth=1)
plt.savefig('part2_spectrum.png')

## Show the plots to see them interactively
plt.show()




import copy
import numpy as np

class Calorimeter:
    '''This defines the calorimeter. The model is a strict one dimensinal model,
    where layers are positioned along the positive z direction and are imagined to
    stretch infinitely into the x and y directions.'''

    class Volume:
        '''A simple volume of the detector that has a layer starting at a given z position'''

        def __init__(self, z, layer):
            self.z = z
            self.layer = layer

    def __init__(self, layers=[]):
        self._layers = layers.copy()
        self._zend = 0

    def add_layer(self, layer):
        '''Add a single layer to the back of the calorimeter.'''
        self._layers.append(self.Volume(self._zend, copy.copy(layer)))
        self._zend += layer._thickness

    def add_layers(self, layers):
        '''Add a list of layers, one after the other to the back of the calorimeter.'''
        for l in layers:
            self.add_layer(l)

    def step(self, particle, step):
        '''Move a particle by the amount step forward in the calorimeter,
        Return a list of particles created during
        the step. If particle doesn't do anything it is just stepped forward.'''

        involume = False
        for volume in self._layers:
            if (particle.z >= volume.z) and (particle.z < volume.z + volume.layer._thickness):
                involume = True
                break

        particle.move(step)

        particles = [particle]
        if involume:
            layer = volume.layer
            layer.ionise(particle, step)
            particles = layer.interact(particle, step)

        return particles

    def positions(self, active=True):
        '''Provide an array of the z coordinates for the start of each layer. If active=True, only return the active layers'''
        return np.array([v.z for v in self._layers if not active or v.layer._yield>0])

    def ionisations(self, active=True):
        '''Provide a list of the ionisation deposited in each of the layers. If active=True, only return the active layers'''
        return np.array([v.layer._ionisation for v in self._layers if not active or v.layer._yield>0])

    def reset(self):
        '''Clears the recorded ionisation in each layer'''
        for v in self._layers:
            v.layer._ionisation=0

    def __str__(self):
        txt = 'The layers of the calorimeter:\n'
        for volume in self._layers:
            txt += f'{volume.z:.2f} ' + str(volume.layer) + '\n'
        return txt



#!/usr/bin/env python
# coding: utf-8

# ------------------------------------------------------------------------------
# Q3: Fourier Analysis
# Part 1: analysing a sound file
# ------------------------------------------------------------------------------

# Import the relevant libraries
import monashspa
import matplotlib.pyplot as plt
import numpy as np
import wave
from matplotlib.widgets import Cursor

# ------------------------------------------------------------------------------
# First we have to import our recorded sample and extract the waveform 
# ------------------------------------------------------------------------------

## Choose a file to plot
filename = "PHS2061_Q3_Guitar_Pluck.wav"

## Open the file
signal_wave = wave.open(filename, 'r')

## Extract the sample rate from the file
sample_rate = signal_wave.getframerate()

## Create a numpy array from the waveform
sig = np.frombuffer(signal_wave.readframes(-1), dtype=np.int16)

## The audio is recorded in stereo so we want to
## split the left and right channels apart.
## We will only use the left channel from here onwards
left, right = sig[0::2], sig[1::2]


# ------------------------------------------------------------------------------
# Now we can plot our sample. 
# Try looking at different ranges by changing the start and end points.
# ------------------------------------------------------------------------------

## If you want to trim the sample 
## choose suitable beginning and end points

# sample_start  = 0
# sample_end    = len(left)
sample_start  = 90000
sample_end    = 200000

## Create the values for the x axis
sample_numbers = np.arange(len(left))

## The variables we want to plot on our graph
x = sample_numbers[sample_start:sample_end]
y = left[sample_start:sample_end]

## Create a plot for the left channel 
fig, (ax0, ax1) = plt.subplots(1, 2)
ax0.plot(x,y)
ax0.set_xlabel('Sample number')
ax0.set_ylabel('Amplitude')
ax0.set_title('Left channel')

## Create a plot for the right channel
x = sample_numbers[sample_start:sample_end]
y = right[sample_start:sample_end]
ax1.plot(x,y)
ax1.set_xlabel('Sample number')
ax1.set_title('Right channel')

fig.suptitle('Recorded waveforms')

plt.savefig('part1_waveform.png')

# ------------------------------------------------------------------------------
# In order to work out what freqencies are present we must do a fourier transform. 
# Luckily NumPy has a function to do this for us.
# ------------------------------------------------------------------------------

## Perform a fast fourier transform of the waveform
data_to_transform = left[sample_start:sample_end]
fourier_transform = np.fft.rfft(data_to_transform)

## Take the absolute value
abs_fourier_transform =  np.abs(fourier_transform)

## Create an array with the frequencies for each point in the transform 
freq = np.fft.rfftfreq(data_to_transform.size, d=1./sample_rate)


# ------------------------------------------------------------------------------
# Now we can try plotting the fourier transform. 
# What do you notice about the positions of the peaks? 
# ------------------------------------------------------------------------------


fig, (ax0) = plt.subplots(1, 1)
ax0.plot(freq,abs_fourier_transform)
ax0.set_xlabel('Frequency (Hz)')
ax0.set_ylabel('Absolute value of FFT')
ax0.set_yscale('log')
ax0.set_title('Fourier transfrom of recorded waveform')

## Pick which range of frequencies to plot
ax0.set_xlim([0, 1000])
cursor = Cursor(ax0, color='green', linewidth=1)
plt.savefig('part1_spectrum.png')

## Show the plots to see them interactively
plt.show()



# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import os
import xml.etree.ElementTree as ET

import numpy as np

def read_data(filepath):
    """Imports the acquired spectrum from the Mossbauer data file.

    Arguments:
        filepath: The path to the .ws5 file produced by the muon
                  Mossbauer acquisition software.

    Returns:
        A 1D numpy array containing the data (counts) from the file. 

    """

    tree = ET.parse(filepath)
    root = tree.getroot()

    data = None

    for child in root:
        if child.tag == 'data':
            channels = int(child.attrib['channels'])
            raw_data = child.text
            split_data = raw_data.split('\n')
            split_data = [x for x in split_data if x]
            if len(split_data) != channels:
                raise RuntimeError('The file {filename} may be malformed. The number of channels ({channels}) specified does not match the number of data points found ({len_data}).'.format(filename=os.path.basename(filepath), channels=channels, len_data=len(split_data)))
            int_data = [int(x) for x in split_data]
            data = np.array(int_data)
    
    if data is None:
        raise RuntimeError('Could not locate the data in the file {filename}. Is your data saved in the correct file format?'.format(filename=os.path.basename(filepath)))
    else:
        return data



#!/usr/bin/env python

import random

import monashspa.PHS3302.calorimeter.model as model
import matplotlib.pyplot as plt
import numpy as np


# ## An example illustrating how the code works
# 
# Create the calorimeter. In this case a sampling calorimeter with interchaging
# layers of scintillator (low density, active) and lead (high density, inactive).

mycal = model.Calorimeter()

lead = model.Layer('lead', 2.0, 0.5, 0.0)
scintillator = model.Layer('Scin', 0.01, 0.5, 1.0)

for i in range(40):
    mycal.add_layers([lead, scintillator])


# Run a simulation


zcors = mycal.positions()

sim = model.Simulation(mycal)
ionisations = sim.simulate(model.Electron(0.0, 10.0), 25)

meanionisations = np.mean(ionisations, axis=0)
rmsionisations = np.std(ionisations, axis=0)
energies = np.sum(ionisations, axis=1)
rel_resolution = np.std(energies)/np.mean(energies)

print(f'Relative resolution is {rel_resolution:.3f}')


# Create some plots

fig = plt.figure(figsize=(15,5))
ax1, ax2, ax3 = fig.subplots(1, 3)
for i in ionisations:
    ax1.plot(zcors, i)
ax1.set_xlabel(r'$x$ [cm]')
ax1.set_ylabel(r'ionisation')

ax2.errorbar(zcors, meanionisations, yerr=rmsionisations)
ax2.set_xlabel(r'$x$ [cm]')
ax2.set_ylabel(r'ionisation (mean and spread)')

ax3.hist(energies)

plt.tight_layout()
plt.show()


# # Your exercise should start here

# ## Study energy resolution
# Make a plot of the relative energy resolution, $\sigma(E)/E$ for particles between 1 and 10 GeV.
# You should be able to demonstrate that the resolution is proportional to $1/\sqrt{E}$.
# This will require that you run the simulation code above inside a loop to get number for multiple
# different energies.

particle_energies = np.logspace(0.0, 1.1, 10)


# 2 marks: Create loop to find resolutions for different energies
# 2 marks: Create plot of the relative energy resolution vs. energy
# 2 marks: Document that it shows the $1/\sqrt{E}$ behaviour through a fit or similar

# ## Noise
# Modify the code such that when reading out the total ionisation from a layer, a noise term is added.
# The noise is a random amount from a Gaussian distribution that is added to the ionisation for active
# layers. Demonstrate how this is only relevant for the relative resolution when the ingoing particle
# has low energy. Identify a noise level yourself that illustrates the effect well.

# 2 marks. Add noise term correctly as an additive term on a per layer basis.
# 2 marks: Create aplot that shows the new relative energy resolution.
# 2 marks: Comparison in one form or the other to situation without the noise.

# ## Calibration
# Study how mis-calibration of layers will affect the resolution. Implement this by scaling the
# total ionisation collected in all layers for a given particle by a random value. The scaling factor
# should be close to 1.0. Demonstrate how this is most relevant for the relative resolution when the
# ingoing particle has high energy.

# 2 marks. Add calibration term correctly as a multiplicative term on the overall ionisation.
# 2 marks: Create a plot that shows the new relative energy resolution.
# 2 marks: Comparison in one form or the other to situation without the mis-calibration.

# ## Fit an overall resolution
# 
# Put everything together, so that you have a simulation that considers both noise and calibration.
# Implement a fit that determines the a, b anc c parameters of a calorimeter resolution model:
# $\frac{\sigma(E)}{E} = \frac{a}{\sqrt{E}} \oplus b \oplus \frac{c}{E}$.


if False:
    # Fill these arrays with the relative resolution and the uncertainty on that estimate for each of the
    # particle energies
    rel_resolutions_withall = []
    u_rel_resolutions_withall = []

    from lmfit import fit_report
    from scipy import stats
    from monashspa.common.fitting import linear_fit, get_fit_parameters, make_lmfit_model, model_fit
    from monashspa.common.figures import savefig

    name='Energy resolution'
    model = make_lmfit_model("sqrt(a*a/x + b*b + c*c/(x*x))")
    params = model.make_params(a=0.09, b=0.03, c=3)
    fit_results = model_fit(model, params, particle_energies, rel_resolutions_withall, u_y=u_rel_resolutions_withall)

    # Extract result and print nicely
    fit = fit_results.best_fit
    u_fit = fit_results.eval_uncertainty(sigma=1)
    fit_parameters = get_fit_parameters(fit_results)
    pvalue = 1.0 - stats.chi2.cdf(fit_results.chisqr,
                                  fit_results.ndata-fit_results.nvarys)

    print("""
    [[{name}]]
    =================
      p-value       = {pvalue:.2E}
    """.format(name=name, pvalue=pvalue))
    print(fit_results.fit_report())

    # Create some plots
    fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
    ax1.errorbar(particle_energies, rel_resolutions_withall, yerr=u_rel_resolutions_withall, marker="x", linestyle="None", color="black",label="Energy resolution")
    ax1.plot(particle_energies, fit, marker="None", linestyle="-", color="black",label="fit to {name}".format(name=name))
    ax1.fill_between(particle_energies, fit-u_fit,fit+u_fit,
                     color="lightgrey",label="uncertainty in {name}".format(name=name))
    ax1.legend(bbox_to_anchor=(0.3,1))
    ax1.set(ylabel="resolution")
    pull = (rel_resolutions_withall-fit)/u_rel_resolutions_withall
    ax2.plot(particle_energies, pull, marker="*", linestyle="None", color="black")
    emin = np.min(particle_energies)
    emax = np.max(particle_energies)
    ax2.plot([emin, emax], [ 0, 0], marker="None", linestyle="-", color="grey")
    ax2.plot([emin, emax], [ 1, 1], marker="None", linestyle="dashed", color="grey")
    ax2.plot([emin, emax], [-1,-1], marker="None", linestyle="dashed", color="grey")
    ax2.set(xlabel="x", ylabel="Pull")
    scale = 1.1*np.max(np.abs(pull))
    ax2.set_ylim(-scale, scale)
    fig.suptitle('Fit with {name}'.format(name=name))

    plt.show()

# 2 marks. Simulate data with both noise and mis-calibration
# 2 marks: Fit the simulated data
# 3 marks: Interpret results by discussing the shape of the graphs and commenting on the fit.



import os
import sys

import numpy as np
import matplotlib.pyplot as plt

from lmfit import fit_report
from scipy import stats
from monashspa.common.fitting import linear_fit, get_fit_parameters, make_lmfit_model, model_fit
from monashspa.common.figures import savefig

xval = np.arange(start=-5.0, stop=5.0, step=0.8)
u_yval = 0.8 + xval*xval/15.
yval = 1.5*xval + np.random.normal(0.0,u_yval)

#######################################################################

# Name of the model that you fit
name = "linear model"

#######################################################################

# Create the model and run fit
model = make_lmfit_model("a*x + b")
params = model.make_params(a=1.0, b=0.0)
fit_results = model_fit(model, params, xval, yval, u_y=u_yval)

# Extract result and print nicely
fit = fit_results.best_fit
u_fit = fit_results.eval_uncertainty(sigma=1)
fit_parameters = get_fit_parameters(fit_results)
pvalue = 1.0 - stats.chi2.cdf(fit_results.chisqr,
                              fit_results.ndata-fit_results.nvarys)

print("""
[[{name}]]
=================
  p-value       = {pvalue:.2E}
""".format(name=name, pvalue=pvalue))
print(fit_results.fit_report())


# Create some plots
fig, ax = plt.subplots(1)
ax.errorbar(xval, yval, yerr=u_yval, marker="x", linestyle="None", color="black",label="experimental data")
ax.plot(xval, fit, marker="None", linestyle="-", color="black",label="fit to {name}".format(name=name))
#ax1.fill_between(xval, fit-u_fit,fit+u_fit,
#                 color="lightgrey",label="uncertainty in {name}".format(name=name))
ax.legend(bbox_to_anchor=(0.6,1))
ax.set(xlabel = "x", ylabel="y")
fig.suptitle('Fit with {name}'.format(name=name))
savefig('chisquare_linear.png'.format(name=name))


# Create some plots
fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(xval, yval, yerr=u_yval, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(xval, fit, marker="None", linestyle="-", color="black",label="fit to {name}".format(name=name))
#ax1.fill_between(xval, fit-u_fit,fit+u_fit,
#                 color="lightgrey",label="uncertainty in {name}".format(name=name))
ax1.legend(bbox_to_anchor=(0.6,1))
ax1.set(ylabel="y")
pull = (yval-fit)/u_yval
ax2.plot(xval, pull, marker="*", linestyle="None", color="black")
ax2.plot([np.min(xval), np.max(xval)], [0,0], marker="None", linestyle="-", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="x", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with {name}'.format(name=name))
savefig('chisquare_linear_pull.png'.format(name=name))



#########################
# Wrong uncertainty scale
#########################

# Underestimate uncertainties in fit
# Create the model and run fit
yval = 1.5*xval + np.random.normal(0.0,u_yval)
model_underestimate = make_lmfit_model("a*x + b")
params_underestimate = model_underestimate.make_params(a=1.0, b=0.0)
fit_results_underestimate = model_fit(model_underestimate, params_underestimate, xval, yval, u_y=u_yval/3.0)


# Extract result and print nicely
fit_underestimate = fit_results_underestimate.best_fit
u_fit = fit_results_underestimate.eval_uncertainty(sigma=1)
fit_parameters = get_fit_parameters(fit_results_underestimate)
pvalue_underestimate = 1.0 - stats.chi2.cdf(fit_results_underestimate.chisqr,
                              fit_results_underestimate.ndata-fit_results_underestimate.nvarys)

print("""
[[{name}]]
=================
  p-value       = {pvalue:.2E}
""".format(name=name, pvalue=pvalue_underestimate))
print(fit_results_underestimate.fit_report())

# Create some plots
fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(xval, yval, yerr=u_yval/3.0, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(xval, fit_underestimate, marker="None", linestyle="-", color="black",
         label="fit to {name}".format(name=name))
#ax1.fill_between(xval, fit-u_fit,fit+u_fit,
#                 color="lightgrey",label="uncertainty in {name}".format(name=name))
ax1.legend(bbox_to_anchor=(0.6,1))
ax1.set(ylabel="y")
pull = (yval-fit_underestimate)/(u_yval/3.0)
ax2.plot(xval, pull, marker="*", linestyle="None", color="black")
ax2.plot([np.min(xval), np.max(xval)], [0,0], marker="None", linestyle="-", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="x", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with {name}'.format(name=name))
savefig('chisquare_linear_underestimate.png'.format(name=name))


###############################
# Wrong model for uncertainties
###############################

# Underestimate uncertainties in fit
# Create the model and run fit
yval = 1.5*xval + np.random.normal(0.0,u_yval)
u_yval_wrong = (xval+6.0)/3.0
model_underestimate = make_lmfit_model("a*x + b")
params_underestimate = model_underestimate.make_params(a=1.0, b=0.0)
fit_results_underestimate = model_fit(model_underestimate, params_underestimate, xval, yval, u_y=u_yval_wrong)


# Extract result and print nicely
fit_underestimate = fit_results_underestimate.best_fit
u_fit = fit_results_underestimate.eval_uncertainty(sigma=1)
fit_parameters = get_fit_parameters(fit_results_underestimate)
pvalue_underestimate = 1.0 - stats.chi2.cdf(fit_results_underestimate.chisqr,
                              fit_results_underestimate.ndata-fit_results_underestimate.nvarys)

print("""
[[{name}]]
=================
  p-value       = {pvalue:.2E}
""".format(name=name, pvalue=pvalue_underestimate))
print(fit_results_underestimate.fit_report())

# Create some plots
fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(xval, yval, yerr=u_yval_wrong, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(xval, fit_underestimate, marker="None", linestyle="-", color="black",
         label="fit to {name}".format(name=name))
#ax1.fill_between(xval, fit-u_fit,fit+u_fit,
#                 color="lightgrey",label="uncertainty in {name}".format(name=name))
ax1.legend(bbox_to_anchor=(0.6,1))
ax1.set(ylabel="y")
pull = (yval-fit_underestimate)/(u_yval_wrong)
ax2.plot(xval, pull, marker="*", linestyle="None", color="black")
ax2.plot([np.min(xval), np.max(xval)], [0,0], marker="None", linestyle="-", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="x", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with {name}'.format(name=name))
savefig('chisquare_linear_wrong_uncertainty.png'.format(name=name))

###############################
# Over fitting 
###############################
# 
# Create the model and run fit
yval = 1.5*xval + np.random.normal(0.0,u_yval)
model_underestimate = make_lmfit_model("a*x*x*x*x + b*x*x*x + c*x*x + d*x + e")
params_underestimate = model_underestimate.make_params(a=0.0, b=0.0, c=0.0, d=1.5, e=0.0)
fit_results_underestimate = model_fit(model_underestimate, params_underestimate, xval, yval, u_y=u_yval)


# Extract result and print nicely
fit_underestimate = fit_results_underestimate.best_fit
u_fit = fit_results_underestimate.eval_uncertainty(sigma=1)
fit_parameters = get_fit_parameters(fit_results_underestimate)
pvalue_underestimate = 1.0 - stats.chi2.cdf(fit_results_underestimate.chisqr,
                              fit_results_underestimate.ndata-fit_results_underestimate.nvarys)

print("""
[Quartic model]]
=================
  p-value       = {pvalue:.2E}
""".format(name=name, pvalue=pvalue_underestimate))
print(fit_results_underestimate.fit_report())

# Create some plots
fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(xval, yval, yerr=u_yval, marker="x", linestyle="None", color="black",label="experimental data")
ax1.plot(xval, fit_underestimate, marker="None", linestyle="-", color="black",
         label="fit to {name}".format(name=name))
#ax1.fill_between(xval, fit-u_fit,fit+u_fit,
#                 color="lightgrey",label="uncertainty in {name}".format(name=name))
ax1.legend(bbox_to_anchor=(0.6,1))
ax1.set(ylabel="y")
pull = (yval-fit_underestimate)/(u_yval)
ax2.plot(xval, pull, marker="*", linestyle="None", color="black")
ax2.plot([np.min(xval), np.max(xval)], [0,0], marker="None", linestyle="-", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot([np.min(xval), np.max(xval)], [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="x", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with quartic function'.format(name=name))
savefig('chisquare_cubic.png'.format(name=name))




plt.show()



#######################################################################
# Fit the data to the ideal gas law
# Assume n=0.040, p=1.013
#
# V = nR/p * T = 0.040 / 1.013 * R * T

# Change the filename to read a different name
datafile = 'PHS2062_gas_short_data.csv'

# Change the model you fit here. "x" takes the role of temperature
model = "0.040/1.013*R*x"

# Name of the model that you fit
name = "Ideal gas law"

#######################################################################

import os
import sys

import pandas

import numpy as np
import matplotlib.pyplot as plt

from lmfit import fit_report
from scipy import stats
from monashspa.common.fitting import linear_fit, get_fit_parameters, make_lmfit_model, model_fit
from monashspa.common.figures import savefig

# Read the data from file
data = None
try:
    folder = os.path.dirname(os.path.realpath(__file__))
    fname = os.path.join(folder, datafile)
    data = pandas.read_csv(fname, usecols=[0,1,2]).to_numpy()
except FileNotFoundError:
    print("ERROR: Failed to read CSV file {file} with data. It should be in the same folder {folder} as the Python file".format(file=datafile, folder=folder))

if type(data)==type(None):
    sys.exit(1)

volume = data[:,0]
u_volume = data[:,1]
temp = data[:,2]

# Create the model and run fit
gas_model = make_lmfit_model(model)
gas_params = gas_model.make_params(R=0.01)
gas_fit_results = model_fit(gas_model, gas_params, temp, volume, u_y=u_volume)


# Extract result and print nicely
gas_fit = gas_fit_results.best_fit
u_gas_fit = gas_fit_results.eval_uncertainty(sigma=1)
gas_fit_parameters = get_fit_parameters(gas_fit_results)
pvalue = 1.0 - stats.chi2.cdf(gas_fit_results.chisqr,
                              gas_fit_results.ndata-gas_fit_results.nvarys)

print("""
[[{name}]]
=================
  p-value       = {pvalue:.2E}
""".format(name=name, pvalue=pvalue))
print(gas_fit_results.fit_report())


# Create some plots
fig, (ax1, ax2) = plt.subplots(2, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
ax1.errorbar(temp, volume, yerr=u_volume, marker="*", linestyle="None", color="black",label="experimental data")
ax1.plot(temp, gas_fit, marker="None", linestyle="-", color="black",label="linear fit")
ax1.fill_between(temp, gas_fit-u_gas_fit,gas_fit+u_gas_fit,
                 color="lightgrey",label="uncertainty in linear fit")
ax1.legend(bbox_to_anchor=(0.5,1))
ax1.set(ylabel="Volume (L)")
pull = (volume-gas_fit)/u_volume
ax2.plot(temp, pull, marker="*", linestyle="None", color="black")
ax2.plot([np.min(temp), np.max(temp)], [0,0], marker="None", linestyle="-", color="grey")
ax2.plot([np.min(temp), np.max(temp)], [1,1], marker="None", linestyle="dashed", color="grey")
ax2.plot([np.min(temp), np.max(temp)], [-1,-1], marker="None", linestyle="dashed", color="grey")
ax2.set(xlabel="Temperature (K)", ylabel="Pull")
scale = 1.1*np.max(np.abs(pull))
ax2.set_ylim(-scale, scale)
fig.suptitle('Fit with {name}'.format(name=name))
savefig('{name}.png'.format(name=name))

plt.show()



# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import os
from setuptools import setup, find_packages

# Define the current version of the library
# Update this before building and publishing a new version
# see https://semver.org/ for guidelines on how to modify the version string
VERSION = '1.9.11'
# version information generated by our GitHub Action (see .github/workflows/release.yml)
CURRENT_TAG = os.environ.get('GHA_TAG_VERSION')
CURRENT_GIT_COMMIT_HASH = os.environ.get('GHA_GIT_COMMIT')
if CURRENT_TAG is not None:
    print('GitHub Action debug: tag=', CURRENT_TAG)
if CURRENT_GIT_COMMIT_HASH is not None:
    print('GitHub Action debug: git hash=', CURRENT_GIT_COMMIT_HASH)
# Modify the current version if we are building inside our GitHub action
if CURRENT_TAG is not None and CURRENT_TAG and CURRENT_TAG != 'refs/heads/master':
    VERSION = CURRENT_TAG
elif CURRENT_GIT_COMMIT_HASH:
    index = 0
    try:
        import requests
        response = requests.get('https://test.pypi.org/pypi/monashspa/json', timeout=2)
        data = response.json()
        if response.status_code == 200:
            found = True
            while found:
                # PyPI converts '-' to '.' in version
                if '{version}.dev{index}'.format(version=VERSION, index=index) not in data['releases']:
                    break
                else:
                    index += 1
        else:
            print('GitHub Action debug: Error while getting JSON data from test PyPI. Status code was:', response.status_code)
    except BaseException:
        print('GitHub Action debug: Error while querying PyPI release information. Will guess -dev0 is the correct postfix.')
        pass

    # Despite PyPI converting '-' to '.', we want to use '-' in __version__.py as it matches semver specs.
    VERSION += '-dev{}'.format(index)

# get directory of setup.py and the rest of the code for the library
code_dir = os.path.abspath(os.path.dirname(__file__))

# Auto generate a __version__ file for the package to import
with open(os.path.join(code_dir, 'monashspa', '__version__.py'), 'w') as f:
    f.write("__version__ = '%s'\n" % VERSION)

# Work around the fact that the readme.md file doesn't exist for users installing
# from the tar.gz format. However, in this case, they won't be uploading to PyPi
# so they don't need it!
try:
    # Read in the readme file as the long description
    with open(os.path.join(code_dir, 'readme.md')) as f:
        long_description = f.read()
except Exception:
    long_description = ""

setup(
    name='monashspa',
    version=VERSION,
    description='Library of useful data analysis tools for Monash University Physics & Astronomy students',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://monashspa.readthedocs.io',
    project_urls={
        "Documentation": "https://monashspa.readthedocs.io",
        "Source Code": "https://github.com/Monash-University-Physics-Astronomy/monashspa",
    },
    license='GPLv3',
    author='School of Physics & Astronomy, Monash University',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 'Programming Language :: Python :: 3.11',
                 'Environment :: Console',
                 'Intended Audience :: Education',
                 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                ],
    python_requires='>=3.8, <4',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy>=1.15',
        'lmfit>=1.0,<2',
        'requests>=2.21.0,<3',
        'colorama>=0.4.1,<1',
        'pandas>=1.0,<2',
        'scipy>=1.4.1,<2',
        'piradon',
        'matplotlib',
        'h5py',
        'importlib-metadata',
        'sympy',
        'uproot4',
    ],
    data_files=[
        ('monashspa/PHS2061', [ 'monashspa/PHS2061/PHS20x1UncertaintiesData.csv',
                                'monashspa/PHS2061/PHS2061_Q3_Guitar_Pluck.wav',
                                'monashspa/PHS2061/PHS2061_Q3_Fourier_Analysis.ipynb',
                                'monashspa/PHS2061/PHS2061_Q3_Fourier_Analysis.py',
                                'monashspa/PHS2061/PHS2061_Q3_Synthesiser.ipynb',
                                'monashspa/PHS2061/PHS2061_Q3_Synthesiser.py' ]),
        ('monashspa/PHS3000/tutorials', ['monashspa/PHS3000/tutorials/PHS3000UncertaintiesData.csv']),
        ('monashspa/PHS2062', ['monashspa/PHS2062/PHS2062_gas_short_data.csv',
                               'monashspa/PHS2062/PHS2062_gas_short_2_data.csv',
                               'monashspa/PHS2062/PHS2062_gas_data.csv',
                               'monashspa/PHS2062/PHS2062_gas_fit.py',
                               'monashspa/PHS2062/PHS2062_gas_fit.ipynb',
        ]),
        ('monashspa/PHS3302/calorimeter', ['monashspa/PHS3302/calorimeter/calorimeter.py',
                                           'monashspa/PHS3302/calorimeter/calorimeter.ipynb',
        ]),
        ('monashspa/PHS3302/d0_lifetime', ['monashspa/PHS3302/d0_lifetime/d0_lifetime.py',
                                           'monashspa/PHS3302/d0_lifetime/d0_lifetime.ipynb',
        ]),
    ]
)
